@@ -8,27 +8,28 @@ package cli
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"os"
-	"runtime"
 	"strings"
 
 	"github.com/containers/image/v5/types"
 	"github.com/spf13/cobra"
-	"github.com/sylabs/singularity/docs"
-	"github.com/sylabs/singularity/internal/pkg/cache"
-	"github.com/sylabs/singularity/internal/pkg/client/library"
-	"github.com/sylabs/singularity/internal/pkg/client/net"
-	"github.com/sylabs/singularity/internal/pkg/client/oci"
-	"github.com/sylabs/singularity/internal/pkg/client/oras"
-	"github.com/sylabs/singularity/internal/pkg/client/shub"
-	"github.com/sylabs/singularity/internal/pkg/runtime/launcher"
-	"github.com/sylabs/singularity/internal/pkg/runtime/launcher/native"
-	ocilauncher "github.com/sylabs/singularity/internal/pkg/runtime/launcher/oci"
-	"github.com/sylabs/singularity/internal/pkg/util/uri"
-	"github.com/sylabs/singularity/pkg/syfs"
-	"github.com/sylabs/singularity/pkg/sylog"
-	useragent "github.com/sylabs/singularity/pkg/util/user-agent"
+	"github.com/sylabs/singularity/v4/docs"
+	"github.com/sylabs/singularity/v4/internal/pkg/cache"
+	"github.com/sylabs/singularity/v4/internal/pkg/client/library"
+	"github.com/sylabs/singularity/v4/internal/pkg/client/net"
+	"github.com/sylabs/singularity/v4/internal/pkg/client/oci"
+	"github.com/sylabs/singularity/v4/internal/pkg/client/oras"
+	"github.com/sylabs/singularity/v4/internal/pkg/client/shub"
+	"github.com/sylabs/singularity/v4/internal/pkg/runtime/launcher"
+	"github.com/sylabs/singularity/v4/internal/pkg/runtime/launcher/native"
+	ocilauncher "github.com/sylabs/singularity/v4/internal/pkg/runtime/launcher/oci"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/uri"
+	"github.com/sylabs/singularity/v4/pkg/ocibundle/ocisif"
+	"github.com/sylabs/singularity/v4/pkg/syfs"
+	"github.com/sylabs/singularity/v4/pkg/sylog"
+	useragent "github.com/sylabs/singularity/v4/pkg/util/user-agent"
 )
 
 const (
@@ -47,12 +48,18 @@ func getCacheHandle(cfg cache.Config) *cache.Handle {
 	return h
 }
 
+type contextKey string
+
+const (
+	keyOrigImageURI contextKey = "origImageURI"
+)
+
 // actionPreRun will:
-//   - run replaceURIWithImage;
 //   - do the proper path unsetting;
 //   - and implement flag inferences for:
 //     --compat
 //     --hostname
+//   - run replaceURIWithImage;
 func actionPreRun(cmd *cobra.Command, args []string) {
 	// For compatibility - we still set USER_PATH so it will be visible in the
 	// container, and can be used there if needed. USER_PATH is not used by
@@ -60,12 +67,13 @@ func actionPreRun(cmd *cobra.Command, args []string) {
 	userPath := strings.Join([]string{os.Getenv("PATH"), defaultPath}, ":")
 	os.Setenv("USER_PATH", userPath)
 
-	replaceURIWithImage(cmd.Context(), cmd, args)
-
 	// --compat infers other options that give increased OCI / Docker compatibility
 	// Excludes uts/user/net namespaces as these are restrictive for many Singularity
 	// installs.
 	if isCompat {
+		if noCompat {
+			sylog.Fatalf("Cannot use --no-compat with --compat: incompatible options")
+		}
 		isContainAll = true
 		isWritableTmpfs = true
 		noInit = true
@@ -77,6 +85,9 @@ func actionPreRun(cmd *cobra.Command, args []string) {
 	if len(hostname) > 0 {
 		utsNamespace = true
 	}
+
+	origImageURI := replaceURIWithImage(cmd.Context(), cmd, args)
+	cmd.SetContext(context.WithValue(cmd.Context(), keyOrigImageURI, &origImageURI))
 }
 
 func handleOCI(ctx context.Context, imgCache *cache.Handle, cmd *cobra.Command, pullFrom string) (string, error) {
@@ -90,6 +101,7 @@ func handleOCI(ctx context.Context, imgCache *cache.Handle, cmd *cobra.Command,
 		OciAuth:    ociAuth,
 		DockerHost: dockerHost,
 		NoHTTPS:    noHTTPS,
+		OciSif:     isOCI,
 	}
 
 	return oci.Pull(ctx, imgCache, pullFrom, pullOpts)
@@ -123,7 +135,16 @@ func handleLibrary(ctx context.Context, imgCache *cache.Handle, pullFrom string)
 	if err != nil {
 		return "", err
 	}
-	return library.Pull(ctx, imgCache, r, runtime.GOARCH, tmpDir, c)
+
+	pullOpts := library.PullOptions{
+		Endpoint:      currentRemoteEndpoint,
+		LibraryConfig: c,
+		// false to allow OCI execution of native SIF from library
+		RequireOciSif: false,
+		TmpDir:        tmpDir,
+		Platform:      getOCIPlatform(),
+	}
+	return library.Pull(ctx, imgCache, r, pullOpts)
 }
 
 func handleShub(ctx context.Context, imgCache *cache.Handle, pullFrom string) (string, error) {
@@ -134,71 +155,63 @@ func handleNet(ctx context.Context, imgCache *cache.Handle, pullFrom string) (st
 	return net.Pull(ctx, imgCache, pullFrom, tmpDir)
 }
 
-func replaceURIWithImage(ctx context.Context, cmd *cobra.Command, args []string) {
-	// If args[0] is not transport:ref (ex. instance://...) formatted return, not a URI
-	t, _ := uri.Split(args[0])
+func replaceURIWithImage(ctx context.Context, cmd *cobra.Command, args []string) string {
+	origImageURI := args[0]
+	t, _ := uri.Split(origImageURI)
+	// If joining an instance (instance://xxx), or we have a bare filename then
+	// no retrieval / conversion is required.
 	if t == "instance" || t == "" {
-		return
+		return origImageURI
 	}
 
 	var image string
 	var err error
 
-	// Create a cache handle only when we know we are are using a URI
+	// Create a cache handle only when we know we are using a URI
 	imgCache := getCacheHandle(cache.Config{Disable: disableCache})
 	if imgCache == nil {
 		sylog.Fatalf("failed to create a new image cache handle")
 	}
 
-	// The OCI runtime launcher will handle OCI image sources directly.
-	if ociRuntime {
-		if oci.IsSupported(t) != t {
-			sylog.Fatalf("OCI runtime only supports OCI image sources. %s is not supported.", t)
-		}
-		return
-	}
-
 	switch t {
 	case uri.Library:
-		image, err = handleLibrary(ctx, imgCache, args[0])
+		image, err = handleLibrary(ctx, imgCache, origImageURI)
 	case uri.Oras:
-		image, err = handleOras(ctx, imgCache, cmd, args[0])
+		image, err = handleOras(ctx, imgCache, cmd, origImageURI)
 	case uri.Shub:
-		image, err = handleShub(ctx, imgCache, args[0])
+		image, err = handleShub(ctx, imgCache, origImageURI)
 	case oci.IsSupported(t):
-		image, err = handleOCI(ctx, imgCache, cmd, args[0])
+		image, err = handleOCI(ctx, imgCache, cmd, origImageURI)
 	case uri.HTTP:
-		image, err = handleNet(ctx, imgCache, args[0])
+		image, err = handleNet(ctx, imgCache, origImageURI)
 	case uri.HTTPS:
-		image, err = handleNet(ctx, imgCache, args[0])
+		image, err = handleNet(ctx, imgCache, origImageURI)
 	default:
 		sylog.Fatalf("Unsupported transport type: %s", t)
 	}
 
+	var mountErr *ocisif.UnavailableError
+	if errors.As(err, &mountErr) {
+		if !canUseTmpSandbox {
+			sylog.Fatalf("OCI-SIF functionality could not be used, and fallback to unpacking OCI bundle in temporary sandbox dir disallowed (original error msg: %s)", err)
+		}
+
+		sylog.Warningf("OCI-SIF functionality could not be used, falling back to unpacking OCI bundle in temporary sandbox dir (original error msg: %s)", err)
+		return origImageURI
+	}
+
 	if err != nil {
-		sylog.Fatalf("Unable to handle %s uri: %v", args[0], err)
+		sylog.Fatalf("Unable to handle %s uri: %v", origImageURI, err)
+	}
+
+	// TODO - drop once we have implemented prefix-less oci-sif vs native sif detection.
+	if isOCI {
+		image = "oci-sif:" + image
 	}
 
 	args[0] = image
-}
 
-// setVM will set the --vm option if needed by other options
-func setVM(cmd *cobra.Command) {
-	// check if --vm-ram or --vm-cpu changed from default value
-	for _, flagName := range []string{"vm-ram", "vm-cpu"} {
-		if flag := cmd.Flag(flagName); flag != nil && flag.Changed {
-			// this option requires the VM setting to be enabled
-			cmd.Flags().Set("vm", "true")
-			return
-		}
-	}
-
-	// since --syos is a boolean, it cannot be added to the above list
-	if isSyOS && !vm {
-		// let the user know that passing --syos implicitly enables --vm
-		sylog.Warningf("The --syos option requires a virtual machine, automatically enabling --vm option.")
-		cmd.Flags().Set("vm", "true")
-	}
+	return origImageURI
 }
 
 // ExecCmd represents the exec command
@@ -209,20 +222,13 @@ var ExecCmd = &cobra.Command{
 	PreRun:                actionPreRun,
 	Run: func(cmd *cobra.Command, args []string) {
 		// singularity exec <image> <command> [args...]
-		image := args[0]
-		containerCmd := "/.singularity.d/actions/exec"
-		containerArgs := args[1:]
-		// OCI runtime does not use an action script
-		if ociRuntime {
-			containerCmd = args[1]
-			containerArgs = args[2:]
+		ep := launcher.ExecParams{
+			Image:   args[0],
+			Action:  "exec",
+			Process: args[1],
+			Args:    args[2:],
 		}
-		setVM(cmd)
-		if vm {
-			execVM(cmd, image, containerCmd, containerArgs)
-			return
-		}
-		if err := launchContainer(cmd, image, containerCmd, containerArgs, ""); err != nil {
+		if err := launchContainer(cmd, ep); err != nil {
 			sylog.Fatalf("%s", err)
 		}
 	},
@@ -241,29 +247,11 @@ var ShellCmd = &cobra.Command{
 	PreRun:                actionPreRun,
 	Run: func(cmd *cobra.Command, args []string) {
 		// singularity shell <image>
-		image := args[0]
-		containerCmd := "/.singularity.d/actions/shell"
-		containerArgs := []string{}
-		// OCI runtime does not use an action script, but must match behavior.
-		// See - internal/pkg/util/fs/files/action_scripts.go (case shell).
-		if ociRuntime {
-			// SINGULARITY_SHELL or --shell has priority
-			if shellPath != "" {
-				containerCmd = shellPath
-				// Clear the shellPath - not handled internally by the OCI runtime, as we exec directly without an action script.
-				shellPath = ""
-			} else {
-				// Otherwise try to exec /bin/bash --norc, falling back to /bin/sh
-				containerCmd = "/bin/sh"
-				containerArgs = []string{"-c", "test -x /bin/bash && PS1='Singularity> ' exec /bin/bash --norc || PS1='Singularity> ' exec /bin/sh"}
-			}
+		ep := launcher.ExecParams{
+			Image:  args[0],
+			Action: "shell",
 		}
-		setVM(cmd)
-		if vm {
-			execVM(cmd, image, containerCmd, containerArgs)
-			return
-		}
-		if err := launchContainer(cmd, image, containerCmd, containerArgs, ""); err != nil {
+		if err := launchContainer(cmd, ep); err != nil {
 			sylog.Fatalf("%s", err)
 		}
 	},
@@ -282,19 +270,12 @@ var RunCmd = &cobra.Command{
 	PreRun:                actionPreRun,
 	Run: func(cmd *cobra.Command, args []string) {
 		// singularity run <image> [args...]
-		image := args[0]
-		containerCmd := "/.singularity.d/actions/run"
-		containerArgs := args[1:]
-		// OCI runtime does not use an action script
-		if ociRuntime {
-			containerCmd = ""
+		ep := launcher.ExecParams{
+			Image:  args[0],
+			Action: "run",
+			Args:   args[1:],
 		}
-		setVM(cmd)
-		if vm {
-			execVM(cmd, args[0], containerCmd, containerArgs)
-			return
-		}
-		if err := launchContainer(cmd, image, containerCmd, containerArgs, ""); err != nil {
+		if err := launchContainer(cmd, ep); err != nil {
 			sylog.Fatalf("%s", err)
 		}
 	},
@@ -313,14 +294,12 @@ var TestCmd = &cobra.Command{
 	PreRun:                actionPreRun,
 	Run: func(cmd *cobra.Command, args []string) {
 		// singularity test <image> [args...]
-		image := args[0]
-		containerCmd := "/.singularity.d/actions/test"
-		containerArgs := args[1:]
-		if vm {
-			execVM(cmd, image, containerCmd, containerArgs)
-			return
+		ep := launcher.ExecParams{
+			Image:  args[0],
+			Action: "test",
+			Args:   args[1:],
 		}
-		if err := launchContainer(cmd, image, containerCmd, containerArgs, ""); err != nil {
+		if err := launchContainer(cmd, ep); err != nil {
 			sylog.Fatalf("%s", err)
 		}
 	},
@@ -331,20 +310,21 @@ var TestCmd = &cobra.Command{
 	Example: docs.RunTestExample,
 }
 
-func launchContainer(cmd *cobra.Command, image string, containerCmd string, containerArgs []string, instanceName string) error {
+func launchContainer(cmd *cobra.Command, ep launcher.ExecParams) error {
 	ns := launcher.Namespaces{
-		User: userNamespace,
-		UTS:  utsNamespace,
-		PID:  pidNamespace,
-		IPC:  ipcNamespace,
-		Net:  netNamespace,
+		User:  userNamespace,
+		UTS:   utsNamespace,
+		PID:   pidNamespace,
+		IPC:   ipcNamespace,
+		Net:   netNamespace,
+		NoPID: noPidNamespace,
 	}
 
 	cgJSON, err := getCgroupsJSON()
 	if err != nil {
 		return err
 	}
-	if cgJSON != "" && strings.HasPrefix(image, "instance://") {
+	if cgJSON != "" && strings.HasPrefix(ep.Image, "instance://") {
 		cgJSON = ""
 		sylog.Warningf("Resource limits & cgroups configuration are only applied to instances at instance start.")
 	}
@@ -388,7 +368,7 @@ func launchContainer(cmd *cobra.Command, image string, containerCmd string, cont
 		launcher.OptCgroupsJSON(cgJSON),
 		launcher.OptConfigFile(configurationFile),
 		launcher.OptShellPath(shellPath),
-		launcher.OptPwdPath(pwdPath),
+		launcher.OptCwdPath(cwdPath),
 		launcher.OptFakeroot(isFakeroot),
 		launcher.OptNoSetgroups(noSetgroups),
 		launcher.OptBoot(isBoot),
@@ -399,12 +379,16 @@ func launchContainer(cmd *cobra.Command, image string, containerCmd string, cont
 		launcher.OptKeyInfo(ki),
 		launcher.OptSIFFuse(sifFUSE),
 		launcher.OptCacheDisabled(disableCache),
+		launcher.OptDevice(device),
+		launcher.OptCdiDirs(cdiDirs),
+		launcher.OptNoCompat(noCompat),
+		launcher.OptNoTmpSandbox(noTmpSandbox),
 	}
 
 	// Explicitly use the interface type here, as we will add alternative launchers later...
 	var l launcher.Launcher
 
-	if ociRuntime {
+	if isOCI {
 		sylog.Debugf("Using OCI runtime launcher.")
 
 		sysCtx := &types.SystemContext{
@@ -432,5 +416,32 @@ func launchContainer(cmd *cobra.Command, image string, containerCmd string, cont
 		}
 	}
 
-	return l.Exec(cmd.Context(), image, containerCmd, containerArgs, instanceName)
+	err = l.Exec(cmd.Context(), ep)
+	var mountErr *ocisif.UnavailableError
+	if !(errors.As(err, &mountErr) && strings.HasPrefix(ep.Image, "oci-sif:")) {
+		return err
+	}
+
+	if !canUseTmpSandbox {
+		return fmt.Errorf("OCI-SIF functionality could not be used, and fallback to unpacking OCI bundle in temporary sandbox dir disallowed (original error msg: %w)", err)
+	}
+
+	sylog.Warningf("OCI-SIF functionality could not be used, falling back to unpacking OCI bundle in temporary sandbox dir (original error msg: %s)", err)
+	// Create a cache handle only when we know we are using a URI
+	imgCache := getCacheHandle(cache.Config{Disable: disableCache})
+	if imgCache == nil {
+		sylog.Fatalf("failed to create a new image cache handle")
+	}
+	origImageURIPtr := cmd.Context().Value(keyOrigImageURI)
+	if origImageURIPtr == nil {
+		return fmt.Errorf("unable to recover original image URI from context while attempting temp-dir OCI fallback (original OCI-SIF err: %w)", mountErr)
+	}
+
+	origImageURI, ok := origImageURIPtr.(*string)
+	if !ok {
+		return fmt.Errorf("unable to recover original image URI (expected string, found: %T) from context while attempting temp-dir OCI fallback (original OCI-SIF err: %w)", origImageURIPtr, mountErr)
+	}
+	ep.Image = *origImageURI
+
+	return l.Exec(cmd.Context(), ep)
 }
